/* --------------------------------------------------------------- */
/* ITG3200.c -- gyroscope device I/O for Tollos                    */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* These functions provide basic IO for using the ITG3200 3-axis   */
/* gyroscope device, attached via I2C.                             */
/*                                                                 */
/* The following connections are assumed:                          */
/*                                                                 */
/* * SCL & SDA ... to I2C bus SCL & SDA, with appropriate pullups  */
/* (channel is selectable)                                         */
/* * CS ... to VCC (+3.3V) [enable I2C mode]                       */
/* * INT ... to a suitable input bit pin                           */
/* * AD0 ... to GND or VCC (I2C address 0x68/0x69)                 */
/* * CLKIN ... to ground                                           */
/* * REGOUT & CPOUT ... capacitors to ground                       */
/* * RESV-G ... to GND                                             */
/* * VCC, VLOGIC, & GND ... to suitable power rails, with bypass   */
/* capacitors                                                      */
/*                                                                 */
/* This code assumes only one ITG3200 device will be used on a     */
/* controller, so some state information can be held (the device   */
/* channel and address).  This simplifies the tollosGyro interface */
/* and allows initialization checking.                             */
/* --------------------------------------------------------------- */
// Adding a way to standby axes could be added later (presumably
// with clock source adjustment)

#define  TOLLOS_IO       // to check signatures only
#include "tollos.h"
#include "ITG3200.h"     // device definitions etc.

// state
static uint gyroAddr=0;                 // i2c device address when initialized
static int  gyroRawTemp=GYRO_NOTEMP;    // latest saved temperature value

/* --------------------------------------------------------------- */
/* gyroGetXYZIO -- read latest x, y, and z angular rates           */
/*                                                                 */
/*   x, y, z -- receive angular rates [in milli-degrees/sec]       */
/*                                                                 */
/*   returns 0 if data read, -1 if the device is not initialized,  */
/*     or -2 if there was an error reading the data (or no data    */
/*     ready)                                                      */
/*                                                                 */
/* Results are returned scaled to milli-degrees/sec; resolution is */
/* 69-70 md/s.                                                     */
/*                                                                 */
/* This should normally be called only after a Data Ready          */
/* interrupt; the interrupt is cleared by calling this.            */
/* --------------------------------------------------------------- */
// The ITG3200 also has die-temperature registers, but these are
// ignored at present.

int gyroGetXYZIO(int *x, int *y, int *z) {
  byte status;           // work
  int  rc;               // ..
  int16 value;           // ..
  byte data[8];          // read buffer TH, TL, XH, XL, YH, YL, ZH, ZL

  if (gyroAddr==0) return -1;  // stopped

  // read the status register; this clears the interrupt
  rc=i2cGetReg(gyroAddr, ITG3200_INT_STATUS, &status);

  // the status should show IS_RAW_DATA_RDY because (with tollosGyro)
  // this is only called by the interrupt handler -- however in
  // practice the status register is 0x00; hence the following check
  // cannot be done
  // if (!(status&IS_RAW_DATA_RDY)) return -2;

  // read the data registers; this would also clear interrupt
  rc|=i2cGetRegs(gyroAddr, ITG3200_TEMP_OUT_H, data, 8);
  if (rc!=0) return -3;            // i2c error

  // for the temperature reading the value is assembled and saved in
  // its raw form
  value=(data[0]<<8) | data[1];    // assemble 16 bits
  gyroRawTemp=value;               // sign-extend

  // assemble the 16-bit XYZ values and scale to milli-degrees/sec
  // 14.375 ulps = 1 deg/sec, so the raw data are in the range +/-
  // 28720; to avoid overflow, calculate using 14.375 = 115/8

  // common macro [(int) propagates sign]
  #define gscale(v) ((1000*(int)(v)*8+((v)<0?-57:+57))/115)

  value=(data[2]<<8) | data[3];    // assemble 16 bits
  *x=gscale(value);                // roun, scale, and assign
  value=(data[4]<<8) | data[5];
  *y=gscale(value);
  value=(data[6]<<8) | data[7];
  *z=gscale(value);

  return 0;
  } // gyroGetXYZIO

/* --------------------------------------------------------------- */
/* gyroInitIO -- configure and initialize gyroscope                */
/*                                                                 */
/*   channel -- I2C channel on which device is to be found         */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* The tollosGyro interface expects:                               */
/* * an update rate of at least 25 Hz (but need not be higher than */
/* that on initialization, to save power)                          */
/* * an interrupt [from pin INT1] when Data Ready is raised.       */
/*                                                                 */
/* The caller should set up the interrupt handler before calling   */
/* this function, or clear Data Ready by calling gyroGetXYZIO.     */
/*                                                                 */
/* All three axes of the gyroscope are enabled.  Current           */
/* consumption should be about 6.5mA.                              */
/* --------------------------------------------------------------- */
int gyroInitIO(uint channel) {
  byte b;                // work
  int  rc;               // ..

  gyroAddr=i2cDevice(channel, ITG3200_ADDR_1);
  gyroRawTemp=GYRO_NOTEMP;

  // check the chip is there [it returns its own address, which is
  // not particularly good indentification]
  rc=i2cGetReg(gyroAddr, ITG3200_WHO_AM_I, &b);
  if (rc!=0 || b!=ITG3200_ADDR_1) return -1;      // hopeless

  // from here, rc is accmulated by ORing, so any negative rc
  // will result in a final negative rc

  // reset device to defaults and low power mode
  rc|=i2cPutReg(gyroAddr, ITG3200_PWR_MGM, PM_H_RESET | PM_SLEEP);

  // Set divider and low-pass filter for 25Hz (also sets full-scale)
  rc|=gyroRateIO(25);

  // set up the interrupt configuration:
  //   ACTL=0 -- active high
  //   OPEN=0 -- push-pull
  //   latch until cleared
  //   any read will clear
  //   interrupt on data ready
  rc|=i2cPutReg(gyroAddr, ITG3200_INT_CFG,
                IC_LATCH_INT_EN | IC_INT_ANYRD_2CLEAR |
                IC_RAW_RDY_EN);

  // leave all three axes enabled, drop out of SLEEP mode, and use X
  // axis for the PLL clock
  rc|=i2cPutReg(gyroAddr, ITG3200_PWR_MGM, CLK_SRC_X);

  return rc;
  } // gyroInitIO

/* --------------------------------------------------------------- */
/* gyroRateIO -- change update rate of gyroscope                   */
/*                                                                 */
/*   rate -- rate requested [Hz, 1..200]                           */
/*                                                                 */
/*   returns 0 if rate changed OK, -1 if error (rate zero or       */
/*     too high, or bus error)                                     */
/*                                                                 */
/* The gyroscope will be set to a slowest available rate that      */
/* is >= the requested rate, to minimize resource usage.  The      */
/* low-pass filter bandwidth is set to match or be the next below  */
/* the rate.                                                       */
/*                                                                 */
/* To ensure working with 100kHz I2C, the maximum rate is limited  */
/* to 200Hz -- but note that this may use significant processor    */
/* bandwidth.                                                      */
/* --------------------------------------------------------------- */
// This sets two register fields: the rate divider and the low-pass
// filter; note that the later affects the former if the 256Hz filter
// is used, because that changes the internal sample rate to 8kHz
// (all the others are 1kHz).
//

int gyroRateIO(uint rate) {
  byte lpf;              // low-pass filter setting to use
  byte div;              // required divider, -1
  int  rc=0;             // return code

  if (rate==0) return -1;     // invalid

  // calculate the divider to use, assuming a 1kHz internal rate, and
  // then calculate back the actual rate that will be used
  if (rate<4) {          // within limits ...
    div=255;             // biggest available [divide by 256]
    rate=4;              // lowest as an int
    }
   else {
    div=(1000/rate)-1;
    rate=1000/(div+1);
    }

  // put divider value (-1) into the divider register
  rc|=i2cPutReg(gyroAddr, ITG3200_SMPLRT_DIV, div);

  // use the low-pass filter that is closest but equal to or nearest
  // lower than the frequency requested, limited to 188Hz so the
  // internal sample rate will always be 1kHz
  if       (rate<10)  lpf=DLPF_5HZ;
   else if (rate<20)  lpf=DLPF_10HZ;
   else if (rate<42)  lpf=DLPF_20HZ;
   else if (rate<98)  lpf=DLPF_42HZ;
   else if (rate<200) lpf=DLPF_98HZ;
   else if (rate<256) lpf=DLPF_188HZ;
   else return -1;

  // Only one full scale is allowed, so no need to read the register
  rc|=i2cPutReg(gyroAddr, ITG3200_DLPF_FS, FS_2000 | lpf);

  return rc;
  } // gyroRateIO

/* --------------------------------------------------------------- */
/* gyroStopIO -- stop and power-down gyroscope                     */
/*                                                                 */
/*   returns 0 if gyroscope stopped OK, negative if an error       */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int gyroStopIO(void) {
  uint da=gyroAddr;           // save
  if (gyroAddr==0) return -1; // already stopped
  gyroAddr=0;                 // always require re-initialize
  gyroRawTemp=GYRO_NOTEMP;    // ..

  // reset device to defaults and low power mode
  return i2cPutReg(da, ITG3200_PWR_MGM, PM_H_RESET | PM_SLEEP);
  } // gyroStopIO

/* --------------------------------------------------------------- */
/* gyroTemperatureIO -- return current gyroscope die temperature   */
/*                                                                 */
/*   returns the latest temperature reading in milli-degrees       */
/*     Celsius, or GYRO_NOTEMP if no reading is available          */
/*                                                                 */
/* Converts the recorded raw temperature saved during the last     */
/* read.                                                           */
/* --------------------------------------------------------------- */
int gyroTemperatureIO(void) {

  if (gyroRawTemp==GYRO_NOTEMP) return GYRO_NOTEMP;    // no data

  // 280 ulps change = 1C
  // the datasheet is a little unclear about whether the -13200 ulp
  // (LSB) offset at 35C results in a value of 35C or 0C at that
  // point: experiment showed it would be 0C

  return (gyroRawTemp+13200)*1000/280 + 35000;
  } // gyroTemperatureIO
