/* --------------------------------------------------------------- */
/* stmBitIO.c -- bit device functions for STM                      */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions take a port (GPIO register) number and a bit    */
/* mask -- all those bits that have a 1 in the mask are affected.  */
/*                                                                 */
/* bitSetIO sets the configuration of a bit; bitIntIO controls     */
/* its interrupt capabilities, and bitExistIO checks whether a     */
/* pin exists in the hardware.                                     */
/* --------------------------------------------------------------- */


#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollosBit.h"

#include "stm32f10x.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_gpio.h"



// number and size of ports (LPC = 3x32, STM = 5x16)
#define PORTS 5
#define PBITS 16

// initialization flags
static flag started[PORTS];        // 1 if ready to use
// clock lookup
static uint const clock[PORTS]={RCC_APB2Periph_GPIOA, RCC_APB2Periph_GPIOB,
    RCC_APB2Periph_GPIOC, RCC_APB2Periph_GPIOD, RCC_APB2Periph_GPIOE};
// port lookup
static GPIO_TypeDef * const gpioaddr[PORTS]={GPIOA, GPIOB, GPIOC, GPIOD, GPIOE};

// array of interrupt handlers; one for each bit
static vector bithandlers[PBITS];  // for bits 0...15 in PORTx

// master bit interrupt handler
static void bitHandler(void);
// local initializer (starts clock)
static void bitInitIO(uint port);

/* --------------------------------------------------------------- */
/* bitExistIO -- test whether a bit or bits exist                  */
/*                                                                 */
/*   port   -- port in which the bit or bits reside                */
/*   mask   -- mask of which bit(s) to test                        */
/*                                                                 */
/*   returns 1 if all the bits in the mask exist in the specified  */
/*     port, or 0 otherwise; 0 is also returned if port is invalid */
/*                                                                 */
/* Note: this reports whether the bit exists in a microcontroller, */
/* not whether it is brought out to a board pin, or whether it     */
/* supports interrupts.                                            */
/*                                                                 */
/* It appears that all ports and pins exist on STM32 devices.      */
/* --------------------------------------------------------------- */
// This could be merged into BitSetIO as another action, but as it
// will be used rarely, perhaps, is better if omittable
int bitExistIO(uint port, uint mask) {
  // the mask check here is heavier checking than usual, but does
  // prevent the 'unused parameter' compiler warning
  if (port>=PORTS || mask&0xFFFF0000) return 0;
  return 1;
  } // bitExistIO

/* --------------------------------------------------------------- */
/* bitHandler -- master handler for bit interrupts                 */
/*                                                                 */
/* All interrupts for bits arrive here; the hendler determines     */
/* which bit caused the interrupt and then calls the individual    */
/* bit handler.                                                    */
/*                                                                 */
/* The procedure is to read the status register at the start.      */
/* For each pin for which either a rising or falling edge          */
/* interrupt has been generated, call the pin's interrupt handler  */
/* (that handler can read the bit to find out which happened, if   */
/* BIT_RISEFALL is in use -- STM32 provides no status register for */
/* this).                                                          */
/*                                                                 */
/* Once all the handlers have been called for this interrupt then  */
/* interrupts that were noted are cleared.                         */
/* --------------------------------------------------------------- */
static void bitHandler(void) {
  uint pend=EXTI->PR;              // note pending EXT interrupts
  int  i;                          // work
  uint bit;                        // ..

  bit=0x01;                        // bit to test
  for (i=0; i<PBITS; i++) {        // just the GPIO lines
    if (pend & bit) {
      // if there's a handler set for the bit, invoke it
      // [ignore spurious interrupts, possibly set up 'manually',
      // or caused by user disabling just before interrupt]
      vector hand=bithandlers[i];  // safe copy
      if (hand) hand();
      }
    bit<<=1;
    } // i
  EXTI->PR=pend;                   // clear those just handled
  } // bitHandler

/* --------------------------------------------------------------- */
/* bitInitIO -- initialize a Bit port                              */
/*                                                                 */
/*   port -- port number to initialize                             */
/*                                                                 */
/* This function is called automatically by other functions here   */
/* if the started flag is 0 for the port.                          */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
static void bitInitIO(uint port) {
  RCC_APB2PeriphClockCmd(clock[port], ENABLE);
  started[port]=1;
  } // bitInitIO

/* --------------------------------------------------------------- */
/* bitIntIO -- effect an interrupt configuration of a bit or bits  */
/*                                                                 */
/*   port    -- port in which the bit or bits reside               */
/*   mask    -- mask of which bit(s) to affect                     */
/*   action  -- one of BIT_RISE, _FALL, _RISEFALL, or _DISABLE     */
/*   handler -- interrupt handler to call, or NULL if _DISABLE     */
/*                                                                 */
/*   returns 0 if sucessful or negative if error (e.g., the bit    */
/*     does not exist or does not support interrupts)              */
/*                                                                 */
/* This first disables all interrupts for the specified bit(s).    */
/* It then enables the pins as specified, unless handler is NULL.  */
/*                                                                 */
/* The bit validity check only tests for whether all the specified */
/* bits are valid for interrupts, not whether they are brought out */
/* to a physical pin (they could still be used for self-wrought    */
/* interrupts).                                                    */
/* --------------------------------------------------------------- */
int bitIntIO(uint port, uint mask, uint action, vector handler) {
  int  i, rc;                      // work
  uint bit;                        // ..

  if (port>=PORTS) return -1;
  if (!started[port]) bitInitIO(port);

  // for STM32, interrupts are mapped to EXTIs independent of port
  // hence calls for the same bit in different ports is an error and
  // has to be checked first before altering any state if setting;
  // at the same time, clear any if disabling [this is safe because
  // bitHandler will ignore requests to a NULL handler]
  bit=0x01;                                  // bit to test
  for (i=0; i<PBITS; i++) {
    if (mask & bit) {
      if (action==BIT_DISABLE) bithandlers[i]=NULL;
       else if (bithandlers[i]!=NULL) return -2;  // handler in use
      }
    bit<<=1;
    } // i

  // disable selected lines for interrupts
  EXTI->IMR&=~mask;
  // clear Rising and Falling edge configuration
  EXTI->RTSR&=~mask;
  EXTI->FTSR&=~mask;
  // clear any pending interrupts corresponding to mask bits
  EXTI->PR=mask;
  if (action==BIT_DISABLE) return 0;         // disable request complete

  // here to set up and enable interrupt(s) on GPIO pin(s)

  // Enable required clock (AFIO is used for EXTI line select)
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_AFIO, ENABLE);
  // (GPIO clock will have been started by bitInitIO, above)

  // save handler address in slots for each affected bits, and also
  // connect the Port & Pin to the corresponding EXTI line
  bit=0x01;                                  // bit to test
  for (i=0; i<PBITS; i++) {
    if (mask & bit) {
      bithandlers[i]=handler;
      GPIO_EXTILineConfig(port, i);          // [takes numbers]
      }
    bit<<=1;
    } // i

  // set master interrupt handler(s) if not yet set up; this will
  // also enable the interrupt at NVIC level
  // [error if already in use by a different handler]
  //
  // The master interrupt handler is one of:
  // EXTI0 ... EXTI4  for bits 0...4
  // EXT5_9           for bits 5...9
  // EXT10_15         for bits 10...15
  //
  // It would be easy to speed this up by having dedicated
  // handlers for EXTI0 through EXTI4; currently assuming that
  // consistent response time and simplicity are better, however

  rc=0;                  // accumulate
  if (mask & 0x0001) rc|=sysVector(EXTI0_IRQn, bitHandler, SYSTEM_BITPRIO);
  if (mask & 0x0002) rc|=sysVector(EXTI1_IRQn, bitHandler, SYSTEM_BITPRIO);
  if (mask & 0x0004) rc|=sysVector(EXTI2_IRQn, bitHandler, SYSTEM_BITPRIO);
  if (mask & 0x0008) rc|=sysVector(EXTI3_IRQn, bitHandler, SYSTEM_BITPRIO);
  if (mask & 0x0010) rc|=sysVector(EXTI4_IRQn, bitHandler, SYSTEM_BITPRIO);
  if (mask & 0x03E0) rc|=sysVector(EXTI9_5_IRQn, bitHandler, SYSTEM_BITPRIO);
  if (mask & 0xFC00) rc|=sysVector(EXTI15_10_IRQn, bitHandler, SYSTEM_BITPRIO);
  // any negatives will have accumulated to a negative
  if (rc<0) return -3;                       // already in use

  // set trigger conditions
  if (action==BIT_RISE || action==BIT_RISEFALL) EXTI->RTSR|=mask;
  if (action==BIT_FALL || action==BIT_RISEFALL) EXTI->FTSR|=mask;

  // finally set interrupt enable requests for the bits
  EXTI->IMR|=mask;

  return 0;
  } // bitIntIO

/* --------------------------------------------------------------- */
/* bitQueryIO -- query the state of an input bit or bits           */
/*                                                                 */
/*   port   -- port in which the bit or bits reside                */
/*   mask   -- mask of which bit(s) to query                       */
/*   bits   -- receives the state of the selected bits (other      */
/*             bits will be 0)                                     */
/*                                                                 */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* The result will be -1 if port is not a valid port number.       */
/* Bits are undefined if the selected bits do not exist.           */
/* Bits may also be undefined or unpredictable if the pin          */
/* corresponding to a bit is in alternative use (e.g., for ADC).   */
/* --------------------------------------------------------------- */
int bitQueryIO(uint port, uint mask, uint *bits) {
  if (port>=PORTS) return -1;
  *bits=GPIO_ReadInputData(gpioaddr[port]) & mask;
  return 0;
  } // bitQueryIO

/* --------------------------------------------------------------- */
/* bitSetIO -- effect an action on a bit or bits                   */
/*                                                                 */
/*   port   -- port in which the bit or bits reside                */
/*   mask   -- mask of which bit(s) to affect                      */
/*   action -- one of BIT_ON, BIT_OFF, BIT_INVERT, or              */
/*             BIT_OPENDRAIN for output, or one of BIT_PULLUP,     */
/*             BIT_TRISTATE, or BIT_PULLDOWN for bit input         */
/*                                                                 */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* BIT_ON, BIT_OFF, or BIT_INVERT set the specified bit(s) to      */
/* output mode then sets the bit(s) as appropriate for the request */
/* (BIT_OFF sets each bit to 0, BIT_ON sets them to 1, BIT_INVERT  */
/* flips each bit).  BIT_OPENDRAIN sets the bit to open drain      */
/* output mode and level high.                                     */
/*                                                                 */
/* The actions BIT_PULLUP, BIT_TRISTATE, and BIT_PULLDOWN set the  */
/* specified bit(s) to input mode with pull-up, neither pull-up    */
/* nor pull-down, or pull-down.                                    */
/*                                                                 */
/* After reset, bits are BIT_TRISTATE.                             */
/*                                                                 */
/* For example:                                                    */
/*                                                                 */
/*   rc=bitSet(PORT1, BIT27, BIT_PULLUP);                          */
/*                                                                 */
/* The result is -1 if port is not a valid port number.            */
/* Bits are undefined if the selected bits do not exist.           */
/* --------------------------------------------------------------- */
static uint16 dirs[PORTS];    // pin directions, 0=input

int bitSetIO(uint port, uint mask, uint action) {
  GPIO_InitTypeDef config;    // configuration structure
  uint bit;                   // work
  int  i;                     // ..
  GPIO_TypeDef *gpio;         // address of port

  if (port>=PORTS) return -1;
  if (!started[port]) bitInitIO(port);

  gpio=gpioaddr[port];        // number -> address
  mask&=0x0000ffff;           // safety clear

  switch(action) {
    case BIT_ON:
    case BIT_OFF:
    case BIT_INVERT:
    case BIT_OPENDRAIN: {
      // is an output request: if necessary, check all the affected
      // pins are outputs and configure them if not
      if ((mask & dirs[port])!=mask     // some mismatch
        || action==BIT_OPENDRAIN) {     // or explicit configuration
        config.GPIO_Speed=GPIO_Speed_50MHz;
        config.GPIO_Mode=GPIO_Mode_Out_PP;
        if (action==BIT_OPENDRAIN) config.GPIO_Mode=GPIO_Mode_Out_OD;
        bit=0x01;
        for (i=0; i<PBITS; i++) {
          if (mask & bit) {             // have affected bit
            config.GPIO_Pin=bit;        // set mask
            GPIO_Init(gpio, &config);
            dirs[port]|=bit;            // is now output
            }
          bit<<=1;
          } // i
        } // mismatch

      // now set or reset the bits if required
      switch(action) {
        case BIT_OFF:
          GPIO_ResetBits(gpio, mask);
          break;

        case BIT_OPENDRAIN:             // start high
        case BIT_ON:
          GPIO_SetBits(gpio, mask);
          break;

        case BIT_INVERT: {
          uint read=GPIO_ReadOutputData(gpio);
          GPIO_ResetBits(gpio, mask&read);   // clear those that were 1
          GPIO_SetBits(gpio, mask&(~read));  // set those that were 0
          break;}

        }
      break;}

    case BIT_PULLUP:
    case BIT_PULLDOWN:
    case BIT_TRISTATE: {
      // setting input mode; always configure as requested
      config.GPIO_Speed=GPIO_Speed_50MHz;
      if (action==BIT_PULLUP) config.GPIO_Mode=GPIO_Mode_IPU;
       else if (action==BIT_PULLDOWN) config.GPIO_Mode=GPIO_Mode_IPD;
       else config.GPIO_Mode=GPIO_Mode_IN_FLOATING;

      // now configure all requested pins
      bit=0x01;
      for (i=0; i<PBITS; i++) {
        if (mask & bit) {          // affected bit
          config.GPIO_Pin=bit;     // mask expected
          GPIO_Init(gpio, &config);
          dirs[port]&=~bit;        // is now input
          }
        bit<<=1;
        } // i
      break;}

    default:        // unknown action
      return -2;
      break;
    } // switch

  return 0;
  } // bitSetIO

